@@ -18,10 +18,17 @@
 
 
 import logging
+from typing import Optional
 from urllib.parse import urlsplit
 
-from .filter_helpers import filter_latest_matching, get_latest_by_timestamp
-from .registry_helpers import get_tags, get_token, verify_v2_capability
+from .filter_helpers import filter_latest_matching
+from .http_helpers import build_authenticated_opener, build_authentication_opener
+from .registry_helpers import (
+    get_latest_by_timestamp,
+    get_tags,
+    get_token,
+    verify_v2_capability,
+)
 
 
 logger = logging.getLogger(__name__)
@@ -29,28 +36,36 @@ logger = logging.getLogger(__name__)
 
 def get_latest_tag(
     image: str,
-    tag_part: str,
+    base_tag: str,
     label: str,
-    auth_url: str = "https://auth.docker.io/token",
+    authentication_url: str = "https://auth.docker.io/token",
     registry_url: str = "https://registry-1.docker.io",
     service: str = "registry.docker.io",
+    username: Optional[str] = None,
+    password: Optional[str] = None,
 ) -> str:
     """
     Parse the latest DD-DeCaF-specific tag of a Docker image from its registry.
 
     Args:
         image (str): The fully specified image name, for example, 'dddecaf/wsgi-base'.
-        tag_part (str): The base part of the tag that you are interested in, for
+        base_tag (str): The base part of the tag that you are interested in, for
             example, 'alpine' will match 'dddecaf/wsgi-base:alpine_2020-04-28_24fe0a0'.
         label (str): The image label that defines the build timestamp, for example,
             'dk.dtu.biosustain.wsgi-base.alpine.build.timestamp'.
-        auth_url (str, optional): The URL from where to retrieve an access token for the
-            registry (the default https://auth.docker.io/token corresponds to
-            Docker Hub).
+        authentication_url (str, optional): The URL from where to retrieve an access
+            token for the registry (the default https://auth.docker.io/token
+            corresponds to Docker Hub).
         registry_url (str, optional): The URL of the registry API
             (default https://registry-1.docker.io).
-        service (str, optional): The URL of the service for which to request an access
+        service (str, optional): The service URI for which to request an access
             token (default registry.docker.io).
+        username (str, optional): A username used for authentication when
+            retrieving an access token. Without a username no authentication will be
+            attempted.
+        password (str, optional): A password used for authentication when
+            retrieving an access token. The password field may itself be a token for
+            Bearer authentication when the username is 'oauth2accesstoken'.
 
     Returns:
         str: The very latest Docker image tag in its entirety.
@@ -61,31 +76,28 @@ def get_latest_tag(
         urllib.error.URLError: In case of problems communicating with the registry.
 
     """
-    parts = urlsplit(auth_url)
-    token = get_token(parts, image, service)
-    headers = {
-        "Accept": "application/vnd.docker.distribution.manifest.v2+json",
-        "Authorization": f"Bearer {token}",
-    }
+    auth_opener = build_authentication_opener(registry_url, username, password)
+    token = get_token(auth_opener, urlsplit(authentication_url), image, service)
+    opener = build_authenticated_opener(token)
     parts = urlsplit(registry_url)
-    verify_v2_capability(parts, headers)
-    tags = get_tags(parts, headers, image)
+    verify_v2_capability(opener, parts)
+    tags = get_tags(opener, parts, image)
     if len(tags) == 0:
         raise RuntimeError(f"The requested image {image} does not have any tags.")
     try:
-        latest_tags = filter_latest_matching(tags, tag_part)
+        latest_tags = filter_latest_matching(tags, base_tag)
     except ValueError:
         raise RuntimeError(
             f"The requested image {image} does not have any tags corresponding to the "
-            f"expected format {tag_part}_<date>_<commit>."
+            f"expected format {base_tag}_<date>_<commit>."
         )
     if len(latest_tags) > 1:
-        latest = get_latest_by_timestamp(parts, headers, image, label, latest_tags)
+        latest = get_latest_by_timestamp(opener, parts, image, label, latest_tags)
     elif len(latest_tags) == 1:
         latest = latest_tags[0]
     else:
         raise RuntimeError(
             f"The requested image {image} does not have any tags corresponding to the "
-            f"expected format {tag_part}_<date>_<commit>."
+            f"expected format {base_tag}_<date>_<commit>."
         )
     return str(latest)
